@@ -19,18 +19,19 @@
 #define DEFAULT_SCALE_METHOD              QgsSymbol::ScaleDiameter
 
 #include "qgis_core.h"
-#include "qgis.h"
+// #include "qgis.h"
 #include <QColor>
 #include <QMap>
 #include <QPointF>
 #include <QSet>
 #include <QDomDocument>
 #include <QDomElement>
+#include <QPainterPath>
 
 #include "qgssymbol.h"
-#include "qgssymbollayerutils.h" // QgsStringMap
 #include "qgsfields.h"
 #include "qgspropertycollection.h"
+#include "qgspainteffect.h"
 
 class QPainter;
 class QSize;
@@ -39,7 +40,10 @@ class QPolygonF;
 class QgsDxfExport;
 class QgsExpression;
 class QgsRenderContext;
-class QgsPaintEffect;
+
+#ifndef SIP_RUN
+typedef QMap<QString, QString> QgsStringMap;
+#endif
 
 /**
  * \ingroup core
@@ -67,8 +71,12 @@ class CORE_EXPORT QgsSymbolLayer
           sipType = sipType_QgsFilledMarkerSymbolLayer;
         else if ( sipCpp->layerType() == "SvgMarker" )
           sipType = sipType_QgsSvgMarkerSymbolLayer;
+        else if ( sipCpp->layerType() == "RasterMarker" )
+          sipType = sipType_QgsRasterMarkerSymbolLayer;
         else if ( sipCpp->layerType() == "VectorField" )
           sipType = sipType_QgsVectorFieldSymbolLayer;
+        else if ( sipCpp->layerType() == "MaskMarker" )
+          sipType = sipType_QgsMaskMarkerSymbolLayer;
         else
           sipType = sipType_QgsMarkerSymbolLayer;
         break;
@@ -101,6 +109,8 @@ class CORE_EXPORT QgsSymbolLayer
           sipType = sipType_QgsGradientFillSymbolLayer;
         else if ( sipCpp->layerType() == "ShapeburstFill" )
           sipType = sipType_QgsShapeburstFillSymbolLayer;
+        else if ( sipCpp->layerType() == "RandomMarkerFill" )
+          sipType = sipType_QgsRandomMarkerFillSymbolLayer;
         else
           sipType = sipType_QgsFillSymbolLayer;
         break;
@@ -134,8 +144,8 @@ class CORE_EXPORT QgsSymbolLayer
       PropertyFillStyle, //!< Fill style (eg solid, dots)
       PropertyJoinStyle, //!< Line join style
       PropertySecondaryColor, //!< Secondary color (eg for gradient fills)
-      PropertyLineAngle, //!< Line angle
-      PropertyLineDistance, //!< Distance between lines
+      PropertyLineAngle, //!< Line angle, or angle of hash lines for hash line symbols
+      PropertyLineDistance, //!< Distance between lines, or length of lines for hash line symbols
       PropertyGradientType, //!< Gradient fill type
       PropertyCoordinateMode, //!< Gradient coordinate mode
       PropertyGradientSpread, //!< Gradient spread mode
@@ -160,6 +170,7 @@ class CORE_EXPORT QgsSymbolLayer
       PropertyPlacement, //!< Line marker placement
       PropertyInterval, //!< Line marker interval
       PropertyOffsetAlongLine, //!< Offset along line
+      PropertyAverageAngleLength, //!< Length to average symbol angles over
       PropertyHorizontalAnchor, //!< Horizontal anchor point
       PropertyVerticalAnchor, //!< Vertical anchor point
       PropertyLayerEnabled, //!< Whether symbol layer is enabled
@@ -169,6 +180,12 @@ class CORE_EXPORT QgsSymbolLayer
       PropertyArrowHeadThickness, //!< Arrow head thickness
       PropertyArrowHeadType, //!< Arrow head type
       PropertyArrowType, //!< Arrow type
+      PropertyOffsetX, //!< Horizontal offset
+      PropertyOffsetY, //!< Vertical offset
+      PropertyPointCount, //!< Point count
+      PropertyRandomSeed, //!< Random number seed
+      PropertyClipPoints, //!< Whether markers should be clipped to polygon boundaries
+      PropertyDensityArea, //<! Density area
     };
 
     /**
@@ -179,10 +196,16 @@ class CORE_EXPORT QgsSymbolLayer
 
     virtual ~QgsSymbolLayer();
 
+    //! QgsSymbolLayer cannot be copied
+    QgsSymbolLayer( const QgsSymbolLayer &other ) = delete;
+
+    //! QgsSymbolLayer cannot be copied
+    QgsSymbolLayer &operator=( const QgsSymbolLayer &other ) = delete;
+
     /**
-     * Returns true if symbol layer is enabled and will be drawn.
-     * \since QGIS 3.0
+     * Returns TRUE if symbol layer is enabled and will be drawn.
      * \see setEnabled()
+     * \since QGIS 3.0
      */
     bool enabled() const { return mEnabled; }
 
@@ -190,8 +213,8 @@ class CORE_EXPORT QgsSymbolLayer
      * Sets whether symbol layer is enabled and should be drawn. Disabled
      * layers are not drawn, but remain part of the symbol and can be re-enabled
      * when desired.
-     * \since QGIS 3.0
      * \see enabled()
+     * \since QGIS 3.0
      */
     void setEnabled( bool enabled ) { mEnabled = enabled; }
 
@@ -208,20 +231,20 @@ class CORE_EXPORT QgsSymbolLayer
     /**
      * Set stroke color. Supported by marker and fill layers.
      * \since QGIS 2.1 */
-    virtual void setStrokeColor( const QColor &color ) { Q_UNUSED( color ); }
+    virtual void setStrokeColor( const QColor &color ) { Q_UNUSED( color ) }
 
     /**
-     * Get stroke color. Supported by marker and fill layers.
+     * Gets stroke color. Supported by marker and fill layers.
      * \since QGIS 2.1 */
     virtual QColor strokeColor() const { return QColor(); }
 
     /**
      * Set fill color. Supported by marker and fill layers.
      * \since QGIS 2.1 */
-    virtual void setFillColor( const QColor &color ) { Q_UNUSED( color ); }
+    virtual void setFillColor( const QColor &color ) { Q_UNUSED( color ) }
 
     /**
-     * Get fill color. Supported by marker and fill layers.
+     * Gets fill color. Supported by marker and fill layers.
      * \since QGIS 2.1 */
     virtual QColor fillColor() const { return QColor(); }
 
@@ -231,18 +254,82 @@ class CORE_EXPORT QgsSymbolLayer
      */
     virtual QString layerType() const = 0;
 
+    /**
+     * Called before a set of rendering operations commences on the supplied render \a context.
+     *
+     * This is always followed by a call to stopRender() after all rendering operations
+     * have been completed.
+     *
+     * Subclasses can use this method to prepare for a set of rendering operations, e.g. by
+     * pre-evaluating paths or images to render, and performing other one-time optimisations.
+     *
+     * \see startFeatureRender()
+     * \see stopRender()
+     */
     virtual void startRender( QgsSymbolRenderContext &context ) = 0;
+
+    /**
+     * Called after a set of rendering operations has finished on the supplied render \a context.
+     *
+     * This is always preceded by a call to startRender() before all rendering operations
+     * are commenced.
+     *
+     * Subclasses can use this method to cleanup after a set of rendering operations.
+     *
+     * \see startRender()
+     * \see stopFeatureRender()
+     */
     virtual void stopRender( QgsSymbolRenderContext &context ) = 0;
 
+    /**
+     * Called before the layer will be rendered for a particular \a feature.
+     *
+     * This is always followed by a call to stopFeatureRender() after the feature
+     * has been completely rendered (i.e. all parts have been rendered).
+     *
+     * The default implementation does nothing.
+     *
+     * \note In some circumstances, startFeatureRender() and stopFeatureRender() may not be called
+     * before a symbol layer is rendered. E.g., when a symbol layer is being rendered in isolation
+     * and not as a result of rendering a feature (for instance, when rendering a legend patch or other
+     * non-feature based shape).
+     *
+     * \see stopFeatureRender()
+     * \see startRender()
+     *
+     * \since QGIS 3.12
+     */
+    virtual void startFeatureRender( const QgsFeature &feature, QgsRenderContext &context );
+
+    /**
+     * Called after the layer has been rendered for a particular \a feature.
+     *
+     * This is always preceded by a call to startFeatureRender() just before the feature
+     * will be rendered.
+     *
+     * The default implementation does nothing.
+     *
+     * \note In some circumstances, startFeatureRender() and stopFeatureRender() may not be called
+     * before a symbol layer is rendered. E.g., when a symbol layer is being rendered in isolation
+     * and not as a result of rendering a feature (for instance, when rendering a legend patch or other
+     * non-feature based shape).
+     *
+     * \see startFeatureRender()
+     * \see stopRender()
+     *
+     * \since QGIS 3.12
+     */
+    virtual void stopFeatureRender( const QgsFeature &feature, QgsRenderContext &context );
+
     /**
      * Shall be reimplemented by subclasses to create a deep copy of the instance.
      */
     virtual QgsSymbolLayer *clone() const = 0 SIP_FACTORY;
 
     virtual void toSld( QDomDocument &doc, QDomElement &element, const QgsStringMap &props ) const
-    { Q_UNUSED( props ); element.appendChild( doc.createComment( QStringLiteral( "SymbolLayerV2 %1 not implemented yet" ).arg( layerType() ) ) ); }
+    { Q_UNUSED( props ) element.appendChild( doc.createComment( QStringLiteral( "SymbolLayerV2 %1 not implemented yet" ).arg( layerType() ) ) ); }
 
-    virtual QString ogrFeatureStyle( double mmScaleFactor, double mapUnitScaleFactor ) const { Q_UNUSED( mmScaleFactor ); Q_UNUSED( mapUnitScaleFactor ); return QString(); }
+    virtual QString ogrFeatureStyle( double mmScaleFactor, double mapUnitScaleFactor ) const { Q_UNUSED( mmScaleFactor ) Q_UNUSED( mapUnitScaleFactor ); return QString(); }
 
     /**
      * Should be reimplemented by subclasses to return a string map that
@@ -258,7 +345,7 @@ class CORE_EXPORT QgsSymbolLayer
      */
     virtual QgsSymbol *subSymbol() { return nullptr; }
 
-    //! set layer's subsymbol. takes ownership of the passed symbol
+    //! Sets layer's subsymbol. takes ownership of the passed symbol
     virtual bool setSubSymbol( QgsSymbol *symbol SIP_TRANSFER ) { delete symbol; return false; }
 
     QgsSymbol::SymbolType type() const { return mType; }
@@ -275,7 +362,7 @@ class CORE_EXPORT QgsSymbolLayer
       drawn with an stroke will draw half the width
       of the stroke outside of the polygon. This amount is estimated, since it may
       be affected by data defined symbology rules.*/
-    virtual double estimateMaxBleed( const QgsRenderContext &context ) const { Q_UNUSED( context ); return 0; }
+    virtual double estimateMaxBleed( const QgsRenderContext &context ) const { Q_UNUSED( context ) return 0; }
 
     /**
      * Sets the units to use for sizes and widths within the symbol layer. Individual
@@ -285,7 +372,7 @@ class CORE_EXPORT QgsSymbolLayer
      * \param unit output units
      * \see outputUnit()
      */
-    virtual void setOutputUnit( QgsUnitTypes::RenderUnit unit ) { Q_UNUSED( unit ); }
+    virtual void setOutputUnit( QgsUnitTypes::RenderUnit unit ) { Q_UNUSED( unit ) }
 
     /**
      * Returns the units to use for sizes and widths within the symbol layer. Individual
@@ -297,12 +384,24 @@ class CORE_EXPORT QgsSymbolLayer
      */
     virtual QgsUnitTypes::RenderUnit outputUnit() const { return QgsUnitTypes::RenderUnknownUnit; }
 
-    virtual void setMapUnitScale( const QgsMapUnitScale &scale ) { Q_UNUSED( scale ); }
+    virtual void setMapUnitScale( const QgsMapUnitScale &scale ) { Q_UNUSED( scale ) }
     virtual QgsMapUnitScale mapUnitScale() const { return QgsMapUnitScale(); }
 
-    // used only with rending with symbol levels is turned on (0 = first pass, 1 = second, ...)
-    void setRenderingPass( int renderingPass ) { mRenderingPass = renderingPass; }
-    int renderingPass() const { return mRenderingPass; }
+    /**
+     * Specifies the rendering pass in which this symbol layer should be rendered.
+     * The lower the number, the lower the symbol will be rendered.
+     * 0: first pass, 1: second pass, ...
+     * Defaults to 0
+     */
+    void setRenderingPass( int renderingPass );
+
+    /**
+     * Specifies the rendering pass in which this symbol layer should be rendered.
+     * The lower the number, the lower the symbol will be rendered.
+     * 0: first pass, 1: second pass, ...
+     * Defaults to 0
+     */
+    int renderingPass() const;
 
     /**
      * Returns the set of attributes referenced by the layer. This includes attributes
@@ -313,52 +412,52 @@ class CORE_EXPORT QgsSymbolLayer
     /**
      * Sets a data defined property for the layer. Any existing property with the same key
      * will be overwritten.
+     * \see dataDefinedProperties()
+     * \see Property
      * \since QGIS 3.0
-     * \see getDataDefinedProperty
-     * \see removeDataDefinedProperty
      */
     virtual void setDataDefinedProperty( Property key, const QgsProperty &property );
 
     //! write as DXF
     virtual bool writeDxf( QgsDxfExport &e, double mmMapUnitScaleFactor, const QString &layerName, QgsSymbolRenderContext &context, QPointF shift = QPointF( 0.0, 0.0 ) ) const;
 
-    //! get line width
+    //! Gets line width
     virtual double dxfWidth( const QgsDxfExport &e, QgsSymbolRenderContext &context ) const;
 
-    //! get offset
+    //! Gets offset
     virtual double dxfOffset( const QgsDxfExport &e, QgsSymbolRenderContext &context ) const;
 
-    //! get color
+    //! Gets color
     virtual QColor dxfColor( QgsSymbolRenderContext &context ) const;
 
-    //! get angle
+    //! Gets angle
     virtual double dxfAngle( QgsSymbolRenderContext &context ) const;
 
-    //! get dash pattern
+    //! Gets dash pattern
     virtual QVector<qreal> dxfCustomDashPattern( QgsUnitTypes::RenderUnit &unit ) const;
 
-    //! get pen style
+    //! Gets pen style
     virtual Qt::PenStyle dxfPenStyle() const;
 
-    //! get brush/fill color
+    //! Gets brush/fill color
     virtual QColor dxfBrushColor( QgsSymbolRenderContext &context ) const;
 
-    //! get brush/fill style
+    //! Gets brush/fill style
     virtual Qt::BrushStyle dxfBrushStyle() const;
 
     /**
      * Returns the current paint effect for the layer.
      * \returns paint effect
-     * \since QGIS 2.9
      * \see setPaintEffect
+     * \since QGIS 2.9
      */
     QgsPaintEffect *paintEffect() const;
 
     /**
      * Sets the current paint effect for the layer.
      * \param effect paint effect. Ownership is transferred to the layer.
-     * \since QGIS 2.9
      * \see paintEffect
+     * \since QGIS 2.9
      */
     void setPaintEffect( QgsPaintEffect *effect SIP_TRANSFER );
 
@@ -372,26 +471,41 @@ class CORE_EXPORT QgsSymbolLayer
 
     /**
      * Returns a reference to the symbol layer's property collection, used for data defined overrides.
+     * \see setDataDefinedProperties()
+     * \see Property
      * \since QGIS 3.0
-     * \see setProperties()
      */
     QgsPropertyCollection &dataDefinedProperties() { return mDataDefinedProperties; }
 
     /**
      * Returns a reference to the symbol layer's property collection, used for data defined overrides.
+     * \see setDataDefinedProperties()
      * \since QGIS 3.0
-     * \see setProperties()
      */
     const QgsPropertyCollection &dataDefinedProperties() const { return mDataDefinedProperties; } SIP_SKIP
 
     /**
      * Sets the symbol layer's property collection, used for data defined overrides.
      * \param collection property collection. Existing properties will be replaced.
-     * \since QGIS 3.0
      * \see properties()
+     * \since QGIS 3.0
      */
     void setDataDefinedProperties( const QgsPropertyCollection &collection ) { mDataDefinedProperties = collection; }
 
+    /**
+     * Returns TRUE if the symbol layer (or any of its sub-symbols) contains data defined properties.
+     *
+     * \since QGIS 3.4.5
+     */
+    virtual bool hasDataDefinedProperties() const;
+
+    /**
+     * Returns masks defined by this symbol layer.
+     * This is a list of symbol layers of other layers that should be occluded.
+     * \since QGIS 3.12
+     */
+    virtual QgsSymbolLayerReferenceList masks() const;
+
   protected:
 
     QgsSymbolLayer( QgsSymbol::SymbolType type, bool locked = false );
@@ -399,15 +513,15 @@ class CORE_EXPORT QgsSymbolLayer
     QgsSymbol::SymbolType mType;
 
     //! True if layer is enabled and should be drawn
-    bool mEnabled;
+    bool mEnabled = true;
 
-    bool mLocked;
+    bool mLocked = false;
     QColor mColor;
-    int mRenderingPass;
+    int mRenderingPass = 0;
 
     QgsPropertyCollection mDataDefinedProperties;
 
-    QgsPaintEffect *mPaintEffect = nullptr;
+    std::unique_ptr< QgsPaintEffect > mPaintEffect;
     QgsFields mFields;
 
     // Configuration of selected symbology implementation
@@ -443,6 +557,10 @@ class CORE_EXPORT QgsSymbolLayer
     //! Property definitions
     static QgsPropertiesDefinition sPropertyDefinitions;
 
+#ifdef SIP_RUN
+    QgsSymbolLayer( const QgsSymbolLayer &other );
+#endif
+
 };
 
 //////////////////////
@@ -472,8 +590,16 @@ class CORE_EXPORT QgsMarkerSymbolLayer : public QgsSymbolLayer
       Bottom, //!< Align to bottom of symbol
     };
 
+    //! QgsMarkerSymbolLayer cannot be copied
+    QgsMarkerSymbolLayer( const QgsMarkerSymbolLayer &other ) = delete;
+
+    //! QgsMarkerSymbolLayer cannot be copied
+    QgsMarkerSymbolLayer &operator=( const QgsMarkerSymbolLayer &other ) = delete;
+
     void startRender( QgsSymbolRenderContext &context ) override;
 
+    void stopRender( QgsSymbolRenderContext &context ) override;
+
     /**
      * Renders a marker at the specified point. Derived classes must implement this to
      * handle drawing the point.
@@ -503,9 +629,9 @@ class CORE_EXPORT QgsMarkerSymbolLayer : public QgsSymbolLayer
      * the marker's rotation and data defined rotation before rendering the symbol, and
      * is usually used for orienting symbols to match a line's angle.
      * \param lineAngle Angle in degrees clockwise from north, valid values are between 0 and 360
-     * \since QGIS 2.9
      * \see setAngle()
      * \see angle()
+     * \since QGIS 2.9
      */
     void setLineAngle( double lineAngle ) { mLineAngle = lineAngle; }
 
@@ -516,13 +642,13 @@ class CORE_EXPORT QgsMarkerSymbolLayer : public QgsSymbolLayer
      * \see setSizeUnit()
      * \see setSizeMapUnitScale()
      */
-    void setSize( double size ) { mSize = size; }
+    virtual void setSize( double size ) { mSize = size; }
 
     /**
      * Returns the symbol size. Units are specified by sizeUnit().
      * \see setSize()
      * \see sizeUnit()
-     * \see sizeUnitMapScale()
+     * \see sizeMapUnitScale()
      */
     double size() const { return mSize; }
 
@@ -669,7 +795,7 @@ class CORE_EXPORT QgsMarkerSymbolLayer : public QgsSymbolLayer
      * \param props symbol layer definition (see properties())
      */
     virtual void writeSldMarker( QDomDocument &doc, QDomElement &element, const QgsStringMap &props ) const
-    { Q_UNUSED( props ); element.appendChild( doc.createComment( QStringLiteral( "QgsMarkerSymbolLayer %1 not implemented yet" ).arg( layerType() ) ) ); }
+    { Q_UNUSED( props ) element.appendChild( doc.createComment( QStringLiteral( "QgsMarkerSymbolLayer %1 not implemented yet" ).arg( layerType() ) ) ); }
 
     void setOutputUnit( QgsUnitTypes::RenderUnit unit ) override;
     QgsUnitTypes::RenderUnit outputUnit() const override;
@@ -688,7 +814,7 @@ class CORE_EXPORT QgsMarkerSymbolLayer : public QgsSymbolLayer
 
     /**
      * Constructor for QgsMarkerSymbolLayer.
-     * \param locked set to true to lock symbol color
+     * \param locked set to TRUE to lock symbol color
      */
     QgsMarkerSymbolLayer( bool locked = false );
 
@@ -753,6 +879,10 @@ class CORE_EXPORT QgsMarkerSymbolLayer : public QgsSymbolLayer
   private:
     static QgsMarkerSymbolLayer::HorizontalAnchorPoint decodeHorizontalAnchorPoint( const QString &str );
     static QgsMarkerSymbolLayer::VerticalAnchorPoint decodeVerticalAnchorPoint( const QString &str );
+
+#ifdef SIP_RUN
+    QgsMarkerSymbolLayer( const QgsMarkerSymbolLayer &other );
+#endif
 };
 
 /**
@@ -762,16 +892,139 @@ class CORE_EXPORT QgsMarkerSymbolLayer : public QgsSymbolLayer
 class CORE_EXPORT QgsLineSymbolLayer : public QgsSymbolLayer
 {
   public:
+
+    //! Options for filtering rings when the line symbol layer is being used to render a polygon's rings.
+    enum RenderRingFilter
+    {
+      AllRings, //!< Render both exterior and interior rings
+      ExteriorRingOnly, //!< Render the exterior ring only
+      InteriorRingsOnly, //!< Render the interior rings only
+    };
+
+    //! QgsLineSymbolLayer cannot be copied
+    QgsLineSymbolLayer( const QgsLineSymbolLayer &other ) = delete;
+
+    //! QgsLineSymbolLayer cannot be copied
+    QgsLineSymbolLayer &operator=( const QgsLineSymbolLayer &other ) = delete;
+
+    void setOutputUnit( QgsUnitTypes::RenderUnit unit ) override;
+    QgsUnitTypes::RenderUnit outputUnit() const override;
+    void setMapUnitScale( const QgsMapUnitScale &scale ) override;
+    QgsMapUnitScale mapUnitScale() const override;
+    void drawPreviewIcon( QgsSymbolRenderContext &context, QSize size ) override;
+    double dxfWidth( const QgsDxfExport &e, QgsSymbolRenderContext &context ) const override;
+
+    /**
+     * Renders the line symbol layer along the line joining \a points, using the given render \a context.
+     * \see renderPolygonStroke()
+     */
     virtual void renderPolyline( const QPolygonF &points, QgsSymbolRenderContext &context ) = 0;
 
+    /**
+     * Renders the line symbol layer along the outline of polygon, using the given render \a context.
+     *
+     * The exterior ring of the polygon is specified in \a points. Optionally, interior
+     * rings are set via the \a rings argument.
+     *
+     * \see renderPolyline()
+     */
     virtual void renderPolygonStroke( const QPolygonF &points, QList<QPolygonF> *rings, QgsSymbolRenderContext &context );
 
+    /**
+     * Sets the \a width of the line symbol layer.
+     *
+     * Calling this method updates the width of the line symbol layer, without
+     * changing the existing width units. It has different effects depending
+     * on the line symbol layer subclass, e.g. for a simple line layer it
+     * changes the stroke width of the line, for a marker line layer it
+     * changes the size of the markers used to draw the line.
+     *
+     * \see width()
+     * \warning Since the width units vary, this method is useful for changing the
+     * relative width of a line symbol layer only.
+     */
     virtual void setWidth( double width ) { mWidth = width; }
+
+    /**
+     * Returns the estimated width for the line symbol layer.
+     *
+     * \warning This returned value is inaccurate if the symbol layer has sub-symbols with
+     * different width units. Use the overload accepting a QgsRenderContext
+     * argument instead for accurate sizes in this case.
+     *
+     * \see setWidth()
+     */
     virtual double width() const { return mWidth; }
 
+    /**
+     * Returns the line symbol layer width, in painter units.
+     *
+     * This method returns an accurate width by calculating the actual rendered
+     * width of the symbol layer using the provided render \a context.
+     *
+     * \see setWidth()
+     *
+     * \since QGIS 3.4.5
+     */
+    virtual double width( const QgsRenderContext &context ) const;
+
+    /**
+     * Returns the line's offset.
+     *
+     * Offset units can be retrieved by calling offsetUnit().
+     *
+     * \see setOffset()
+     * \see offsetUnit()
+     * \see offsetMapUnitScale()
+     */
     double offset() const { return mOffset; }
+
+    /**
+     * Sets the line's \a offset.
+     *
+     * Offset units are set via setOffsetUnit().
+     *
+     * \see offset()
+     * \see setOffsetUnit()
+     * \see setOffsetMapUnitScale()
+     */
     void setOffset( double offset ) { mOffset = offset; }
 
+    /**
+     * Sets the \a unit for the line's offset.
+     * \see offsetUnit()
+     * \see setOffset()
+     * \see setOffsetMapUnitScale()
+    */
+    void setOffsetUnit( QgsUnitTypes::RenderUnit unit ) { mOffsetUnit = unit; }
+
+    /**
+     * Returns the units for the line's offset.
+     * \see setOffsetUnit()
+     * \see offset()
+     * \see offsetMapUnitScale()
+    */
+    QgsUnitTypes::RenderUnit offsetUnit() const { return mOffsetUnit; }
+
+    /**
+     * Sets the map unit \a scale for the line's offset.
+     * \see offsetMapUnitScale()
+     * \see setOffset()
+     * \see setOffsetUnit()
+    */
+    void setOffsetMapUnitScale( const QgsMapUnitScale &scale ) { mOffsetMapUnitScale = scale; }
+
+    /**
+     * Returns the map unit scale for the line's offset.
+     * \see setOffsetMapUnitScale()
+     * \see offset()
+     * \see offsetUnit()
+    */
+    const QgsMapUnitScale &offsetMapUnitScale() const { return mOffsetMapUnitScale; }
+
+    // TODO QGIS 4.0 - setWidthUnit(), widthUnit(), setWidthUnitScale(), widthUnitScale()
+    // only apply to simple line symbol layers and do not belong here.
+
     /**
      * Sets the units for the line's width.
      * \param unit width units
@@ -789,30 +1042,28 @@ class CORE_EXPORT QgsLineSymbolLayer : public QgsSymbolLayer
     const QgsMapUnitScale &widthMapUnitScale() const { return mWidthMapUnitScale; }
 
     /**
-     * Sets the units for the line's offset.
-     * \param unit offset units
-     * \see offsetUnit()
-    */
-    void setOffsetUnit( QgsUnitTypes::RenderUnit unit ) { mOffsetUnit = unit; }
+     * Returns the line symbol layer's ring filter, which controls which rings are
+     * rendered when the line symbol is being used to draw a polygon's rings.
+     *
+     * This setting has no effect when the line symbol is not being rendered
+     * for a polygon.
+     *
+     * \see setRingFilter()
+     * \since QGIS 3.6
+     */
+    RenderRingFilter ringFilter() const;
 
     /**
-     * Returns the units for the line's offset.
-     * \see setOffsetUnit()
-    */
-    QgsUnitTypes::RenderUnit offsetUnit() const { return mOffsetUnit; }
-
-    void setOffsetMapUnitScale( const QgsMapUnitScale &scale ) { mOffsetMapUnitScale = scale; }
-    const QgsMapUnitScale &offsetMapUnitScale() const { return mOffsetMapUnitScale; }
-
-    void setOutputUnit( QgsUnitTypes::RenderUnit unit ) override;
-    QgsUnitTypes::RenderUnit outputUnit() const override;
-
-    void setMapUnitScale( const QgsMapUnitScale &scale ) override;
-    QgsMapUnitScale mapUnitScale() const override;
-
-    void drawPreviewIcon( QgsSymbolRenderContext &context, QSize size ) override;
-
-    double dxfWidth( const QgsDxfExport &e, QgsSymbolRenderContext &context ) const override;
+     * Sets the line symbol layer's ring \a filter, which controls which rings are
+     * rendered when the line symbol is being used to draw a polygon's rings.
+     *
+     * This setting has no effect when the line symbol is not being rendered
+     * for a polygon.
+     *
+     * \see ringFilter()
+     * \since QGIS 3.6
+     */
+    void setRingFilter( QgsLineSymbolLayer::RenderRingFilter filter );
 
   protected:
     QgsLineSymbolLayer( bool locked = false );
@@ -823,6 +1074,13 @@ class CORE_EXPORT QgsLineSymbolLayer : public QgsSymbolLayer
     double mOffset = 0;
     QgsUnitTypes::RenderUnit mOffsetUnit = QgsUnitTypes::RenderMillimeters;
     QgsMapUnitScale mOffsetMapUnitScale;
+
+    RenderRingFilter mRingFilter = AllRings;
+
+  private:
+#ifdef SIP_RUN
+    QgsLineSymbolLayer( const QgsLineSymbolLayer &other );
+#endif
 };
 
 /**
@@ -832,6 +1090,13 @@ class CORE_EXPORT QgsLineSymbolLayer : public QgsSymbolLayer
 class CORE_EXPORT QgsFillSymbolLayer : public QgsSymbolLayer
 {
   public:
+
+    //! QgsFillSymbolLayer cannot be copied
+    QgsFillSymbolLayer( const QgsFillSymbolLayer &other ) = delete;
+
+    //! QgsFillSymbolLayer cannot be copied
+    QgsFillSymbolLayer &operator=( const QgsFillSymbolLayer &other ) = delete;
+
     virtual void renderPolygon( const QPolygonF &points, QList<QPolygonF> *rings, QgsSymbolRenderContext &context ) = 0;
 
     void drawPreviewIcon( QgsSymbolRenderContext &context, QSize size ) override;
@@ -845,6 +1110,11 @@ class CORE_EXPORT QgsFillSymbolLayer : public QgsSymbolLayer
     void _renderPolygon( QPainter *p, const QPolygonF &points, const QList<QPolygonF> *rings, QgsSymbolRenderContext &context );
 
     double mAngle = 0.0;
+
+  private:
+#ifdef SIP_RUN
+    QgsFillSymbolLayer( const QgsFillSymbolLayer &other );
+#endif
 };
 
 class QgsSymbolLayerWidget;  // why does SIP fail, when this isn't here
