@@ -18,11 +18,11 @@
 #ifndef QGSWKBTYPES_H
 #define QGSWKBTYPES_H
 
+#include <QObject>
 #include <QMap>
 #include <QString>
 
 #include "qgis_core.h"
-#include "qgis.h"
 
 /***************************************************************************
  * This class is considered CRITICAL and any change MUST be accompanied with
@@ -39,6 +39,7 @@
 
 class CORE_EXPORT QgsWkbTypes
 {
+    Q_GADGET
   public:
 
     /**
@@ -127,6 +128,7 @@ class CORE_EXPORT QgsWkbTypes
       MultiLineString25D,
       MultiPolygon25D
     };
+    Q_ENUM( Type )
 
     /**
      * The geometry types are used to group QgsWkbTypes::Type in a
@@ -142,11 +144,13 @@ class CORE_EXPORT QgsWkbTypes
       UnknownGeometry,
       NullGeometry
     };
+    Q_ENUM( GeometryType )
 
     /**
      * Returns the single type for a WKB type. For example, for MultiPolygon WKB types the single type would be Polygon.
      * \see isSingleType()
      * \see multiType()
+     * \see curveType()
      * \see flatType()
      */
     static Type singleType( Type type )
@@ -291,6 +295,7 @@ class CORE_EXPORT QgsWkbTypes
      * Returns the multi type for a WKB type. For example, for Polygon WKB types the multi type would be MultiPolygon.
      * \see isMultiType()
      * \see singleType()
+     * \see curveType()
      * \see flatType()
      */
     static Type multiType( Type type )
@@ -418,11 +423,271 @@ class CORE_EXPORT QgsWkbTypes
       return Unknown;
     }
 
+
+    /**
+     * Returns the curve type for a WKB type. For example, for Polygon WKB types the curve type would be CurvePolygon.
+     *
+     * \note Returns `CompoundCurve` for `CircularString` (and its Z/M variants)
+     *
+     * \see linearType()
+     * \see isMultiType()
+     * \see isCurvedType()
+     * \see singleType()
+     * \see flatType()
+     * \see multiType()
+     *
+     * \since QGIS 3.10
+     */
+    static Type curveType( Type type )
+    {
+      switch ( type )
+      {
+        case Unknown:
+        case Triangle:
+        case TriangleZ:
+        case TriangleM:
+        case TriangleZM:
+          return Unknown;
+
+        case GeometryCollection:
+          return GeometryCollection;
+
+        case GeometryCollectionZ:
+          return GeometryCollectionZ;
+
+        case GeometryCollectionM:
+          return GeometryCollectionM;
+
+        case GeometryCollectionZM:
+          return GeometryCollectionZM;
+
+        case Point:
+          return Point;
+
+        case MultiPoint:
+          return MultiPoint;
+
+        case PointZ:
+          return PointZ;
+
+        case MultiPointZ:
+          return MultiPointZ;
+
+        case PointM:
+          return PointM;
+
+        case MultiPointM:
+          return MultiPointM;
+
+        case PointZM:
+          return PointZM;
+
+        case MultiPointZM:
+          return MultiPointZM;
+
+        case LineString:
+        case CompoundCurve:
+        case CircularString:
+          return CompoundCurve;
+
+        case MultiLineString:
+        case MultiCurve:
+          return MultiCurve;
+
+        case LineStringZ:
+        case CompoundCurveZ:
+        case CircularStringZ:
+        case LineString25D:
+          return CompoundCurveZ;
+
+        case MultiLineStringZ:
+        case MultiCurveZ:
+        case MultiLineString25D:
+          return MultiCurveZ;
+
+        case LineStringM:
+        case CompoundCurveM:
+        case CircularStringM:
+          return CompoundCurveM;
+
+        case MultiLineStringM:
+        case MultiCurveM:
+          return MultiCurveM;
+
+        case LineStringZM:
+        case CompoundCurveZM:
+        case CircularStringZM:
+          return CompoundCurveZM;
+
+        case MultiLineStringZM:
+        case MultiCurveZM:
+          return MultiCurveZM;
+
+        case Polygon:
+        case CurvePolygon:
+          return CurvePolygon;
+
+        case MultiPolygon:
+        case MultiSurface:
+          return MultiSurface;
+
+        case PolygonZ:
+        case CurvePolygonZ:
+        case Polygon25D:
+          return CurvePolygonZ;
+
+        case MultiPolygonZ:
+        case MultiSurfaceZ:
+        case MultiPolygon25D:
+          return MultiSurfaceZ;
+
+        case PolygonM:
+        case CurvePolygonM:
+          return CurvePolygonM;
+
+        case MultiPolygonM:
+        case MultiSurfaceM:
+          return MultiSurfaceM;
+
+        case PolygonZM:
+        case CurvePolygonZM:
+          return CurvePolygonZM;
+
+        case MultiPolygonZM:
+        case MultiSurfaceZM:
+          return MultiSurfaceZM;
+
+        case NoGeometry:
+          return NoGeometry;
+
+        case Point25D:
+        case MultiPoint25D:
+          return MultiPoint25D;
+      }
+      return Unknown;
+    }
+
+    /**
+     * Returns the linear type for a WKB type. For example, for a CompoundCurve, the linear type would be LineString.
+     *
+     * \see curveType()
+     * \see isMultiType()
+     * \see isCurvedType()
+     * \see singleType()
+     * \see flatType()
+     * \see multiType()
+     *
+     * \since QGIS 3.14
+     */
+    static Type linearType( Type type )
+    {
+      switch ( type )
+      {
+
+        case CircularString:
+        case CompoundCurve:
+          return LineString;
+
+        case CircularStringM:
+        case CompoundCurveM:
+          return LineStringM;
+
+        case CircularStringZ:
+        case CompoundCurveZ:
+          return LineStringZ;
+
+        case CircularStringZM:
+        case CompoundCurveZM:
+          return LineStringZM;
+
+        case MultiCurve:
+          return MultiLineString;
+
+        case MultiCurveM:
+          return MultiLineStringM;
+
+        case MultiCurveZ:
+          return MultiLineStringZ;
+
+        case MultiCurveZM:
+          return MultiLineStringZM;
+
+        case CurvePolygon:
+          return Polygon;
+
+        case CurvePolygonM:
+          return PolygonM;
+
+        case CurvePolygonZ:
+          return PolygonZ;
+
+        case CurvePolygonZM:
+          return PolygonZM;
+
+        case MultiSurface:
+          return MultiPolygon;
+
+        case MultiSurfaceM:
+          return MultiPolygonM;
+
+        case MultiSurfaceZ:
+          return MultiPolygonZ;
+
+        case MultiSurfaceZM:
+          return MultiPolygonZM;
+
+        case GeometryCollection:
+        case GeometryCollectionM:
+        case GeometryCollectionZ:
+        case GeometryCollectionZM:
+        case LineString:
+        case LineString25D:
+        case LineStringM:
+        case LineStringZ:
+        case LineStringZM:
+        case MultiLineString:
+        case MultiLineString25D:
+        case MultiLineStringM:
+        case MultiLineStringZ:
+        case MultiLineStringZM:
+        case MultiPoint:
+        case MultiPoint25D:
+        case MultiPointM:
+        case MultiPointZ:
+        case MultiPointZM:
+        case MultiPolygon:
+        case MultiPolygon25D:
+        case MultiPolygonM:
+        case MultiPolygonZ:
+        case MultiPolygonZM:
+        case NoGeometry:
+        case Point:
+        case Point25D:
+        case PointM:
+        case PointZ:
+        case PointZM:
+        case Polygon:
+        case Polygon25D:
+        case PolygonM:
+        case PolygonZ:
+        case PolygonZM:
+        case Triangle:
+        case TriangleM:
+        case TriangleZ:
+        case TriangleZM:
+        case Unknown:
+          return type;
+
+      }
+      return Unknown;
+    }
+
     /**
      * Returns the flat type for a WKB type. This is the WKB type minus any Z or M dimensions.
      * For example, for PolygonZM WKB types the single type would be Polygon.
      * \see singleType()
      * \see multiType()
+     * \see curveType()
      */
     static Type flatType( Type type )
     {
@@ -540,7 +805,7 @@ class CORE_EXPORT QgsWkbTypes
     static Type parseType( const QString &wktStr );
 
     /**
-     * Returns true if the WKB type is a single type.
+     * Returns TRUE if the WKB type is a single type.
      * \see isMultiType()
      * \see singleType()
      */
@@ -550,7 +815,7 @@ class CORE_EXPORT QgsWkbTypes
     }
 
     /**
-     * Returns true if the WKB type is a multi type.
+     * Returns TRUE if the WKB type is a multi type.
      * \see isSingleType()
      * \see multiType()
      */
@@ -600,7 +865,7 @@ class CORE_EXPORT QgsWkbTypes
     }
 
     /**
-     * Returns true if the WKB type is a curved type or can contain curved geometries.
+     * Returns TRUE if the WKB type is a curved type or can contain curved geometries.
      * \since QGIS 2.14
      */
     static bool isCurvedType( Type type )
@@ -761,7 +1026,7 @@ class CORE_EXPORT QgsWkbTypes
 
     /**
      * Tests whether a WKB type contains the z-dimension.
-     * \returns true if type has z values
+     * \returns TRUE if type has z values
      * \see addZ()
      * \see hasM()
      */
@@ -811,7 +1076,7 @@ class CORE_EXPORT QgsWkbTypes
 
     /**
      * Tests whether a WKB type contains m values.
-     * \returns true if type has m values
+     * \returns TRUE if type has m values
      * \see addM()
      * \see hasZ()
      */
@@ -963,32 +1228,6 @@ class CORE_EXPORT QgsWkbTypes
         return Unknown;
     }
 
-  private:
-
-    struct wkbEntry
-    {
-      wkbEntry( const QString &name, bool isMultiType, Type multiType, Type singleType, Type flatType, GeometryType geometryType,
-                bool hasZ, bool hasM )
-        : mName( name )
-        , mIsMultiType( isMultiType )
-        , mMultiType( multiType )
-        , mSingleType( singleType )
-        , mFlatType( flatType )
-        , mGeometryType( geometryType )
-        , mHasZ( hasZ )
-        , mHasM( hasM )
-      {}
-      QString mName;
-      bool mIsMultiType;
-      Type mMultiType;
-      Type mSingleType;
-      Type mFlatType;
-      GeometryType mGeometryType;
-      bool mHasZ;
-      bool mHasM;
-    };
-
-    static const QMap<Type, wkbEntry> ENTRIES;
 };
 
 #endif // QGSWKBTYPES_H
