@@ -21,10 +21,6 @@ __author__ = 'Victor Olaya'
 __date__ = 'August 2012'
 __copyright__ = '(C) 2012, Victor Olaya'
 
-# This will get replaced with a git SHA1 when you do a git archive
-
-__revision__ = '9b176802e5512bc310765d8ebe4e44b5f4663453'
-
 import os
 import re
 
@@ -35,7 +31,9 @@ from qgis.core import (QgsApplication,
                        QgsProcessingFeatureSourceDefinition,
                        QgsProcessingAlgorithm,
                        QgsProcessingContext,
-                       QgsProcessingFeedback)
+                       QgsProcessingFeedback,
+                       QgsProviderRegistry,
+                       QgsDataSourceUri)
 
 from processing.algs.gdal.GdalAlgorithmDialog import GdalAlgorithmDialog
 from processing.algs.gdal.GdalUtils import GdalUtils
@@ -63,7 +61,7 @@ class GdalAlgorithm(QgsProcessingAlgorithm):
         return self.__class__()
 
     def createCustomParametersWidget(self, parent):
-        return GdalAlgorithmDialog(self)
+        return GdalAlgorithmDialog(self, parent=parent)
 
     def flags(self):
         return QgsProcessingAlgorithm.FlagSupportsBatch # cannot cancel!
@@ -91,6 +89,7 @@ class GdalAlgorithm(QgsProcessingAlgorithm):
                 # and extract selection if required
                 ogr_data_path = self.parameterAsCompatibleSourceLayerPath(parameters, parameter_name, context,
                                                                           QgsVectorFileWriter.supportedFormatExtensions(),
+                                                                          QgsVectorFileWriter.supportedFormatExtensions()[0],
                                                                           feedback=feedback)
                 ogr_layer_name = GdalUtils.ogrLayerName(ogr_data_path)
             else:
@@ -100,22 +99,34 @@ class GdalAlgorithm(QgsProcessingAlgorithm):
                 ogr_data_path = 'path_to_data_file'
                 ogr_layer_name = 'layer_name'
         elif input_layer.dataProvider().name() == 'ogr':
-            if executing:
+            if executing and (isinstance(parameters[parameter_name], QgsProcessingFeatureSourceDefinition) and parameters[parameter_name].selectedFeaturesOnly) \
+                    or input_layer.subsetString():
                 # parameter is a vector layer, with OGR data provider
                 # so extract selection if required
                 ogr_data_path = self.parameterAsCompatibleSourceLayerPath(parameters, parameter_name, context,
                                                                           QgsVectorFileWriter.supportedFormatExtensions(),
                                                                           feedback=feedback)
-                ogr_layer_name = GdalUtils.ogrLayerName(input_layer.dataProvider().dataSourceUri())
+                parts = QgsProviderRegistry.instance().decodeUri('ogr', ogr_data_path)
+                ogr_data_path = parts['path']
+                if 'layerName' in parts and parts['layerName']:
+                    ogr_layer_name = parts['layerName']
+                else:
+                    ogr_layer_name = GdalUtils.ogrLayerName(ogr_data_path)
             else:
+                #either not using the selection, or
                 #not executing - don't worry about 'selected features only' handling. It has no meaning
                 #for the command line preview since it has no meaning outside of a QGIS session!
                 ogr_data_path = GdalUtils.ogrConnectionStringAndFormatFromLayer(input_layer)[0]
                 ogr_layer_name = GdalUtils.ogrLayerName(input_layer.dataProvider().dataSourceUri())
+        elif input_layer.dataProvider().name().lower() == 'wfs':
+            uri = QgsDataSourceUri(input_layer.source())
+            baseUrl = uri.param('url').split('?')[0]
+            ogr_data_path = "WFS:{}".format(baseUrl)
+            ogr_layer_name = uri.param('typename')
         else:
             # vector layer, but not OGR - get OGR compatible path
             # TODO - handle "selected features only" mode!!
-            ogr_data_path = GdalUtils.ogrConnectionString(input_layer.dataProvider().dataSourceUri(), context)
+            ogr_data_path = GdalUtils.ogrConnectionStringFromLayer(input_layer)
             ogr_layer_name = GdalUtils.ogrLayerName(input_layer.dataProvider().dataSourceUri())
         return ogr_data_path, ogr_layer_name
 
@@ -136,16 +147,6 @@ class GdalAlgorithm(QgsProcessingAlgorithm):
 
         return results
 
-    def helpUrl(self):
-        helpPath = GdalUtils.gdalHelpPath()
-        if helpPath == '':
-            return None
-
-        if os.path.exists(helpPath):
-            return QUrl.fromLocalFile(os.path.join(helpPath, '{}.html'.format(self.commandName()))).toString()
-        else:
-            return helpPath + '{}.html'.format(self.commandName())
-
     def commandName(self):
         parameters = {}
         for param in self.parameterDefinitions():
